<!DOCTYPE html>

<html>
	<head>
        <meta charset="utf-8" />
		<title>Psychology Experiment</title>
		<link rel="stylesheet" href="/static/css/bootstrap.min.css" type="text/css" />
		<link rel="stylesheet" href="/static/css/style.css" type="text/css" />
	</head>
	<body>
		<div id="container-not-an-ad" class="media">
			<div class="media-left pull-left" href="#">
						<!-- REPLACE THE LOGO HERE WITH YOUR  UNIVERSITY, LAB, or COMPANY -->
						<img id="adlogo" src="/static/images/university.png" alt="Lab Logo" />
			</div>
			<div class="media-body">
						<h1>Thanks for your participation!</h1>
						<hr />
					    <div class="alert alert-warning">
					    	Our records suggest you have already completed this task.  Thank you.
					    	You cannot complete the same experiment twice.
					    </div>
			</div>
		</div>
	</body>
</html>
