<!DOCTYPE html>

<html>
<head>
        <meta charset="utf-8" />
	<title>Custom Routes</title>
	<link rel="stylesheet" href="/static/css/bootstrap.min.css" type="text/css" />
	<link rel="stylesheet" href="/static/css/style.css" type="text/css" media="screen" />
</head>
<body>
	<div style="text-align: center;" id="thanks">
		<h1>Booyah, this is an example custom route you can add to your project</h1>
	</div>
</body>

</html>
